//----------------------------------------------------------------------------
// Borland Class Library
// Copyright (c) 1991, 2000 by Inprise Corp., All Rights Reserved
//
//$Revision:   1.5  $
//
// Implements commonly used template functions Min(), Max(), range()
//
//   T Min( T, T ) returns the lesser of its arguments
//
//   T Min( T, T, T ) returns the least of its arguments
//
//   T Max( T, T ) returns the greater of its arguments
//
//   T Max( T, T, T ) returns the greatest of its arguments
//
//   T range( T minVal, T maxVal, T val ) returns val if val is
//       between minVal and maxVal.  If val is greater than maxVal,
//       returns maxVal.  If val is less than minVal, returns minVal.
//
//   void Swap( T, T ) swaps the contents of the two given references
//
//----------------------------------------------------------------------------
#if !defined(CLASSLIB_STDTEMPL_H)
#include <classlib/pre.h> /* - pre-CLASSLIB - */
#define CLASSLIB_STDTEMPL_H

#if !defined(CLASSLIB_DEFS_H)
# include <classlib/defs.h>
#endif

//#pragma option -Vo-
#if defined( BI_CLASSLIB_NO_po )
# pragma option -po-
#endif

#if defined(BI_NAMESPACE)
namespace ClassLib {
#endif


//
// The following Min() and Max() are to avoid various Min() and Max()
// compatibility problems
//
template <class T> inline T Min(const T& a, const T& b)
{
  return a<b ? a : b;
}

template <class T> inline T Max(const T& a, const T& b)
{
  return a>b ? a : b;
}

template <class T> inline const T& Min( const T& t1, const T& t2, const T& t3 )
{
    return t1>t2 ? (t2>t3 ? t3 : t2) : (t1>t3 ? t3 : t1 );
}

template <class T> inline T Max( const T& t1, const T& t2, const T& t3 )
{
    return t1>t2 ? (t1>t3 ? t1 : t3) : (t2>t3 ? t2 : t3);
}


template <class T> inline const T& range( const T& minVal, const T& maxVal,
                                          const T& val )
{
    return Min( maxVal, Max( minVal, val ) );
}

template <class T> inline void Swap(T _BIDSFARDATA& a, T _BIDSFARDATA& b)
{
  T t = a;
  a = b;
  b = t;
}

#if defined(BI_NAMESPACE)
}   // namespace ClassLib
#endif

#if defined( BI_CLASSLIB_NO_po )
# pragma option -po.
#endif
//#pragma option -Vo.

#include <classlib/post.h> /* - post-CLASSLIB - */
#endif  // CLASSLIB_STDTEMPL_H
